// module clsocket.cpp
// TCP/IP communication via sockets
//
// This class uses a very simple protocol: each request is answered
// by an acknowledgment by the server
// Requests and achnowledgments are messages defined by class MSGSOCKET
// A message on the line consists of a length information (6 ASCII bytes)
// and a data segment. The first byte of the data segment is an operation byte
//
// Private Copyright Dr.E.Huckert 10-1998
//
// Versions: 17-10-2001 new: timeout in basic_receive()
//                      timeout value set via setsockopt in open()
//           12-12-2002 made ready for binary data
//           2003-05-01 added set_content_length() in recv()
//           2003-06-08 function set_timeout() added
//                      Code for timeout revised (was wrong!)
//           2003-09-15 error handling revised in basic_receive()
           
#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#ifdef WIN32
#include <io.h>
#include <winsock.h>
#else
#include <unistd.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <poll.h>
#define  SOCKET int 
#define  LPSTR  char *
#define  LPHOSTENT  struct hostent * 
#define  LPSOCKADDR struct sockaddr_in * 
#define  INVALID_SOCKET -1
#endif
#include "clsocket.h"
#include "msgsocket.h"

// Constructor
CLIENT_SOCKET::CLIENT_SOCKET()
{
  timeout_ms  = 10000L;
  handle      = INVALID_SOCKET;
  initialized = 0;
  logging     = 0;
#ifdef WIN32
  if (WSAStartup(0x0101, &(this->WSAdata)) != 0)
  {
    this->wsa_ok = 0;
    if (this->logging)
      cerr << "cannot start WSA..." << endl;
  }
  this->wsa_ok = 1;
#endif
}
 
// Destructor
CLIENT_SOCKET::~CLIENT_SOCKET()
{
  if (this->initialized)
    this->close();
#ifdef WIN32
  if (this->wsa_ok)
    WSACleanup();
#endif
}

// sleep a number of milliseconds
int CLIENT_SOCKET::milli_sleep(int millis)
{
int ret;
  ret = 0;
#ifdef WIN32
  Sleep(millis);
#else
#ifdef NANOSLEEP
struct timespec ts1, ts2; 
  //cout << "vor nanosleep()" << endl;
  ts1.tv_sec =  0;
  ts1.tv_nsec = 1000000 * (long)millis;  
  ret = nanosleep(&ts1, &ts2);
  //cout << "nach nanosleep()" << endl;
#else
//pollfd_t pfd[1];
struct pollfd pfd[1];
  pfd[0].fd      = 0;
  pfd[0].events  = 0;
  pfd[0].revents = 0;
  ret = poll(pfd, 1, millis);  
#endif
#endif
  return ret;
}   // end milli_sleep()

// convert a binary IP address (IPV4) to a string
// Warning: this is not reentrant!
char *CLIENT_SOCKET::convert_binary_ip(char *binstr)
{
int n, ic;
char xbuf[8];
static char retbuf[32];
  retbuf[0] = 0;
  for (n=0; n < 4; n++)
  {
    ic = binstr[n] & 0xff;
    sprintf((LPSTR)xbuf, "%0d", ic);
    strcat(retbuf, xbuf);
    if (n < 3) strcat(retbuf, ".");
  }
  return retbuf;
}   // end CLIENT_SOCKET::convert_binary_ip()

void CLIENT_SOCKET::set_timeout(int secs)
{
  if (logging > 1)
    cout << "in CLIENT_SOCKET::set_timeout():" << secs << endl;
#ifdef WIN32
  // Windows: set a timeout value
  /*
    // this seems to be wrong
    struct timeval tv;
    tv.tv_sec  = secs;
    tv.tv_usec = 0;
    setsockopt(this->handle, IPPROTO_TCP, SO_RCVTIMEO, 
               (const char *)&tv, (size_t)(sizeof(tv)));
  */
  int tv;
  tv = secs * 1000L;
  setsockopt(this->handle, SOL_SOCKET, SO_RCVTIMEO, 
             (const char *)&tv, sizeof(tv));
#endif
}   // end set_timeout()

// Create acoket, start a connection to a server
// Returns: the handle for that connection
SOCKET CLIENT_SOCKET::open(const char *host_name,
                           int  port_number)
{
struct sockaddr_in  server_sock_addr;
LPHOSTENT           host_ent;
int                 ret;
unsigned long       hostaddr;
SOCKET              result;
long                argp;
int                 optval;
  this->handle = socket(AF_INET, SOCK_STREAM, 0);
  if (logging > 1)
    cout << "after call to socket(): handle=" << this->handle << endl;
  if (this->handle < 0)
    return -2;

  // Clear and set server address structure 
  host_ent = gethostbyname(host_name);                   
  if (host_ent == NULL)
  {
    cout << "ERROR: gethostbyname() failed for host=" << host_name << endl;
    return -3;
  }
  else
  {
    char **pptr;
    char addr[64];
    char *caddr;
    //
    // get the IP adresses and store them in this->inetaddr
    for (pptr = host_ent->h_addr_list; *pptr != NULL; pptr++)
    {
      caddr = this->convert_binary_ip(*pptr); 
      //sprintf((LPSTR)buf,"\taddress: %s\n", caddr);
      strcpy(this->inetaddr, caddr);
    }
  }
  optval = 1;
  setsockopt(this->handle, IPPROTO_TCP, TCP_NODELAY, 
             (const char *)&optval, (size_t)(sizeof(optval)));
  //
  this->set_timeout(TIMEOUT_SECS);
  //
  server_sock_addr.sin_family      = AF_INET;
  server_sock_addr.sin_port        = htons(port_number);
  server_sock_addr.sin_addr.s_addr = inet_addr(inetaddr);
  ret = connect(this->handle, 
                (struct  sockaddr *)&server_sock_addr, 
                sizeof(struct sockaddr_in));
  if (ret < 0)
  {
    cout << "ERROR: connect() fails" << endl;
    return -4;
  }
  this->initialized = 1;
  if (logging)
    cout << "CLIENT_SOCKET::start() ok" << endl;
  //
  return this->handle;
}  // end CLIENT_SOCKET::open() 

void CLIENT_SOCKET::close()
{
#ifdef WIN32
  if (this->handle >= 0)
       closesocket(this->handle);
#endif
  this->handle = -1;
}   /* end CLIENT_SOCKET::close() */

void CLIENT_SOCKET::set_logging(int flag)
{
  this->logging = flag;
}

// send the required number of bytes
// returns: the number of bytes that have been sent
//          < 0 if error
int CLIENT_SOCKET::basic_send(char *buf, int lbuf)
{
int m, n, nb, anz, ret;
int sleep_count;

  if (logging >= 2)
    cout << "basic_send():should send buffer with length=" << lbuf;
  sleep_count = 0;
  m   = 0;
  nb  = 0;
  anz = lbuf;
  while (nb < lbuf)
  {
    /* problems under Windows...
    ret = this->test_socket();
    if (ret)
    {
      // this is a severe error
      if (this->logging)
        cout << "\tCLIENT_SOCKET::basic_send returns Error -1 for socket " 
             << sockfd << endl;
      return -1;
    }
    */
    //ret = ::send(this->handle, (char *)(&buf[nb]), anz, 0);
#ifdef WIN32
    ret = ::send(this->handle, (char *)(&buf[nb]), anz, 0);
#else
    //ret = ::Send(this->handle, (char *)(&buf[nb]), anz, MSG_NOSIGNAL);
    ret = ::send(this->handle, (char *)(&buf[nb]), anz, 0);
#endif
    if (ret == 0)
    {
      if (this->logging)
        cout << "TIMEOUT in CLIENT_SOCKET::basic_send()" << endl;
      milli_sleep(2);
      if (++sleep_count >= this->timeout_ms)
        break;
      continue;
    }
    if (ret < 0)
    {
      // this is a severe error
      if (this->logging)
        cout << "\tCLIENT_SOCKET::basic_send returns Error -2 for socket " 
             << this->handle << endl;
      return -2;
    }
    //
    // success here
    if (logging >= 2)
      cout << "\tbasic_send():" << ret << " bytes sent" << endl;
    nb  += ret;
    anz -= ret;
  }  // end while ...
  if (nb != lbuf)
  {
    if (this->logging)
      cout << "\tCLIENT_SOCKET::basic_send returns Error -3" << endl;
    return -3;
  }
  return nb;
}   // end CLIENT_SOCKET::basic_send()

// send an event via a socket
// returns: the number of bytes sent
//          < 0 if error
int CLIENT_SOCKET::send(MSGSOCKET *pmsg,
                        int wait_flag)
{
int  ret, nb;
long ll;
char *pbuf;
char xbuf[20];
char length[LENGTH_LENGTH + 1];
  //
  // send the length (LENGTH bytes) first
  ll = pmsg->get_message_length();
  sprintf(xbuf, "%c0%dlu", '%', LENGTH_LENGTH);  
  sprintf(length, xbuf, (long)ll);
  nb = 0;
  ret = this->basic_send(length, LENGTH_LENGTH);
  if (ret <= 0)
  {
    cout << "Error: send over socket:" << ret << endl;
    return -1;
  }
  nb += ret;
  //
  // send the operation byte + the content
  pbuf = new char[ll];
  pbuf[0] = pmsg->get_operation();
  memcpy(pbuf + 1, pmsg->get_content(), pmsg->get_content_length()); 
  ret = this->basic_send(pbuf, ll);
  if (ret < 0)
  {
    cout << "Error: send over socket:" << ret << endl;
    return -3;
  }
  if (this->logging > 1)
    cout << "\tCLIENT_SOCKET::send() OK" << endl;
  delete pbuf;
  nb += ret;
  return nb;
}   // end CLIENT_SOCKET::send()

// returns: the number of received bytes
//          < 0 if error
//          0 if timeout (if get_all == 0)
int CLIENT_SOCKET::basic_receive(char *rbuf, 
                                 int lrbuf,
                                 int  get_all)
{
int ret, nb;
  if (logging > 1)
    cout << "in CLIENT_SOCKET::basic_receive():" << lrbuf << " bytes" << endl;
  nb = 0;
  while (lrbuf > 0)
  {
#ifdef WIN32
    ret = ::recv(this->handle, &rbuf[nb], lrbuf, 0);
#else
    ret = ::recv(this->handle, &rbuf[nb], lrbuf, MSG_WAITALL);
    //ret = ::recv(this->handle, &rbuf[nb], lrbuf, MSG_NOSIGNAL);
#endif
    if (logging > 1)
      cout << "\trecv():" << ret << " bytes" << endl;
    if (ret < 0)
    {
      if (nb > 0)
        break;
      nb = ret;
      break;
    }
    if (ret == 0)
    {
      // probably a timeout
      if (get_all == 0)
        break;
    }
    nb += ret;
    lrbuf -= ret;
    if (ret > 0)
    {
      if (get_all != 0) break;
    }
  }
  return nb;
}   // end CLIENT_SOCKET::basic_receive()

// Reads a LF terminated line via TCP/IP
// The returned line includes all received chars
// returns: the number of received bytes
//          < 0 if error
//          0 if timeout 
int CLIENT_SOCKET::receive_line(char *rbuf, 
                                 int lrbuf)
{
int ret, nb;
char ch;
  if (logging > 1)
    cout << "in CLIENT_SOCKET::receive_line():" <<  lrbuf << " bytes requested" 
         << endl;
  nb      = 0;
  rbuf[0] = 0;
  //
  while (lrbuf > 1)
  {
#ifdef WIN32
    ret = ::recv(this->handle, &ch, 1, 0);
#else
    ret = ::recv(this->handle, &ch, 1, MSG_WAITALL);
#endif
    if (logging > 1)
    {
      if ((ch & 0xff) >= ' ')
        cout << "\trecv():" << ret << " Byte=" << ch << endl;
      else
        cout << "\trecv():" << ret << " Byte=" << (int)ch << endl;
    }  
    if (ret < 0)
    {
      if (nb > 0)
        break;
      nb = -1;
      break;
    }
    if (ret == 0)
    {
      // probably a timeout
      break;
    }
    rbuf[nb++] = ch;
    lrbuf--;
    if (ch == '\n')
      // terminating LF
      break;
  }
  //
  if (nb >= 0)
    rbuf[nb] = 0; 
  return nb;
}   // end CLIENT_SOCKET::receive_line()

// Receive a message via a socket
// Fill the message parts into a CLIENT_SOCKET structure
// Structure of a message:
//   length: LENGTH_LENGTH ASCII Bytes (excluding length)
//   message body: field structured, length depending on length transmitted
//                 before
// Returns: a message object (class MSGSOCKET, must be freed externally) 
//          NULL if error
MSGSOCKET *CLIENT_SOCKET::recv()
{
MSGSOCKET  *pmsg;
long        lmsg;
char        *read_buf;
int         nbytes, nb, ret, sleep_count, anz;
char        rbuf[LENGTH_LENGTH + 1];
  // 
  // get the first part (LENGTH_LENGTH bytes fixed length)
  ret = this->basic_receive(rbuf, LENGTH_LENGTH, 1);
  if (ret != LENGTH_LENGTH)
  {
    if (this->logging > 1)
    cout << "Error: wrong number of bytes for fixed length:" << ret << endl;
    return NULL;
  }
  rbuf[ret] = 0;
  nbytes = ret;
  //
  lmsg = atol(rbuf);
  if (this->logging >= 2)
    cout << "::recv() 1:" << "fixed length=" << rbuf << endl;
  if (lmsg < 0L)
  {
    cout << "Error: received message length not OK" << endl;
    return NULL;
  }
  //
  // receive the second message part
  // put the received data in a MSGSOCKET object
  pmsg = new MSGSOCKET('G', NULL, lmsg + 2);
  if (pmsg == NULL)
  {
    cout << "Error allocation receive buffer" << endl;
    return NULL;
  }
  read_buf = pmsg->get_content();
  //
  sleep_count = 0;
  anz = lmsg;
  nb  = 0;
  while (nb < lmsg)
  {
    ret = this->basic_receive(&read_buf[nb], anz, 1);
    if (ret < 0) 
      // must be an error or EOF
      break;
    if (ret == 0)
    {
      if (++sleep_count >= this->timeout_ms)
        break;
      if (this->logging > 1)
        cout << "TIMEOUT in CLIENT_SOCKET::recv()" << endl;
      milli_sleep(2);
    }
    else
    {
      if (logging > 1)
      {
        cout << "CLIENT_SOCKET::recv():" << ret << " bytes received nb=" 
             << nb << " anz=" << anz << endl; 
      }
      nb += ret;
      anz -= ret;
    }
  }   // while nb < lmsg
  //
  if (nb != lmsg)
  {
    cout << 
      "Error: wrong number of bytes second message part" << endl;
    delete read_buf;
    return NULL;
  }
  read_buf[lmsg] = 0;
  pmsg->set_content_length(lmsg);   // added 2003-05-01
  nbytes += nb;
  //
  return pmsg;
}   // end CLIENT_SOCKET::recv()
