// module srvsocket.cpp
// implements the class  SRVSOCKET 
// purpose: server TCP/IP socket communication
//
// This class uses a very simple protocol: each request is answered
// by an acknowledgment by the server
// Requests and achnowledgments are messages defined by class MSGSOCKET
// A message on the line consists of a length information (6 ASCII bytes)
// and a data segment. The first byte of the data segment is an operation byte
// Copyright Dr.E.Huckert 10-1998

// Versions:
//   2003-05-01 added distinction binary-non binary in send(MSGSOCKET *...)

#include "srvsocket.h"
#include "msgsocket.h"

#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <fcntl.h>
#include <errno.h>
#ifdef WIN32
#include <winsock.h>
#include <io.h>
#else
#include <signal.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <sys/msg.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <sys/resource.h>
#include <unistd.h>
#include <poll.h>
#endif
//using namespace std;

#include "srvsocket.h"
#ifdef WIN32
#include <winsock.h>
#endif

// -----------------------------------------------------------
// Implementation of class SRVSOCKET
// Constructor
void SRVSOCKET::basic_init()
{
  this->timeout_ms      = 10000L;  // milli seconds = 10 seconds
  this->sleep_unit      = 100;      // milli seconds
  this->listen_sockfd   = INVALID_SOCKET;
  this->rw_sockfd       = INVALID_SOCKET;
  this->logging         = 0;
  this->partial         = 0;
}

// Constructor
SRVSOCKET::SRVSOCKET()
{
  this->basic_init();
#ifdef WIN32
  if (WSAStartup(0x0101, &(this->WSAdata)) != 0)
  {
    this->wsa_ok = 0;
    if (this->logging)
      cerr << "cannot start WSA..." << endl;
  }
  this->wsa_ok = 1;
#endif
}

// Constructor : this constructor assumes that the two server sockets
// have been established and that the Windows initialisation for
// the socket system has been run
SRVSOCKET::SRVSOCKET(SOCKET ls,
                     SOCKET rws)
{
  this->basic_init();
  this->listen_sockfd   = ls;
  this->rw_sockfd       = rws;
  this->partial         = 1;
#ifdef WIN32
  this->wsa_ok = 1;
#endif
}

// Destructor
SRVSOCKET::~SRVSOCKET()
{
  if (! this->partial)
  {
    if (this->listen_sockfd != INVALID_SOCKET)
      this->close();
#ifdef WIN32
    if (wsa_ok)
      WSACleanup();
#endif
  }
} 

/* problems under Windows...
// test if there is something on the socket
// Returns: == 0 if nothing happened on the socket
//           <>0 if the socket  has received some bytes or has an error
int SRVSOCKET::test_socket()
{
int ret;
fd_set inp_set;
fd_set exc_set;
struct timeval tv;
  FD_ZERO(&inp_set);
  FD_SET(this->sockfd, &inp_set);
  FD_ZERO(&exc_set);
  FD_SET(this->sockfd, &exc_set);
  //
  tv.tv_sec  = 5;
  tv.tv_usec = 0;
  ret =  select(this->sockfd + 1, &inp_set, NULL, &exc_set, &tv);
  if (ret)
  {
    if (FD_ISSET(this->sockfd, &inp_set))
      return this->sockfd;
    if (FD_ISSET(this->sockfd, &exc_set))
      return INVALID_SOCKET;
  }
  return 0;
}   // end SRVSOCKET::test_socket()
*/

// sleep a number of milliseconds
int SRVSOCKET::milli_sleep(int millis)
{
int ret;
  ret = 0;
#ifdef WIN32
  Sleep(millis);
#else
#ifdef NANOSLEEP
struct timespec ts1, ts2; 
  //cout << "vor nanosleep()" << endl;
  ts1.tv_sec =  0;
  ts1.tv_nsec = 1000000 * (long)millis;  
  ret = nanosleep(&ts1, &ts2);
  //cout << "nach nanosleep()" << endl;
#else
//pollfd_t pfd[1];
struct pollfd pfd[1];
  pfd[0].fd      = 0;
  pfd[0].events  = 0;
  pfd[0].revents = 0;
  ret = poll(pfd, 1, millis);  
#endif
#endif
  return ret;
}   // end milli_sleep()

// returns: the number of received bytes
//          < 0 if error
//          ==0 if timeout
long SRVSOCKET::basic_receive(char *rbuf, long lrbuf)
{
long  sleep_count;
long  ret, nb;
  nb          = 0;
  sleep_count = 0;
  while (lrbuf > 0)
  {
#ifdef WIN32
    ret = ::recv(this->rw_sockfd, &rbuf[nb], lrbuf, 0);
#else
    ret = ::recv(this->rw_sockfd, &rbuf[nb], lrbuf, MSG_WAITALL);
    //ret = ::recv(this->rw_sockfd, &rbuf[nb], lrbuf, MSG_NOSIGNAL);
#endif
    if (ret < 0)
    {
#ifdef WIN32
      int nb;
      nb = WSAGetLastError();
      if (logging)
        cout << "wsock32 error code=" << nb << endl;
#endif
      nb = ret;
      break;
    }
    if (ret == 0)
    {
      sleep_count += this->sleep_unit;
      if (sleep_count > this->timeout_ms)
        break;
      this->milli_sleep(this->sleep_unit);       
    }
    if (ret > 0)
    {
      nb += ret;
      lrbuf -= ret;
    }
  }   // end while ...
  return nb;
}   // end SRVSOCKET::basic_receive()

// Receive a message via a socket
// Fill the message parts into a SRVSOCKET structure
// Structure of a message:
//   length: LENGTH_LENGTH ASCII Bytes (excluding length)
//   message body: field structured, length depending on length transmitted
//                 before
// Returns: a message object (class MSGSOCKET) 
//          NULL if error
MSGSOCKET *SRVSOCKET::recv()
{
MSGSOCKET  *pmsg;
long    nb, lmsg, ret;
int     optval;
char    *mbuf;
int     nbytes, sleep_count, anz;
char    rbuf[LENGTH_LENGTH + 1];
  if (this->logging)
    cout << "in SRVSOCKET::recv()" << endl;
  // 
  while (this->rw_sockfd == INVALID_SOCKET)
  {
    this->rw_sockfd = accept(this->listen_sockfd, NULL, NULL);
    if (this->rw_sockfd < 0)
    {
      if (this->logging > 1)
        cout << "accept() timeout" << endl;
      continue; 
    }
    if (this->logging > 1)
      cout << "accept ok: sockfd=" << this->listen_sockfd 
           << " this->rw_sockfd=" << this->rw_sockfd <<endl;
    optval = 1;
    ::setsockopt(this->rw_sockfd, IPPROTO_TCP, TCP_NODELAY, 
                 (const char *)&optval, (size_t)(sizeof(optval)));
    break;
  }   // end while 1
  //
  // get the first part (LENGTH_LENGTH bytes fixed length)
  ret = this->basic_receive(rbuf, (long)LENGTH_LENGTH);
  if (ret != (long)LENGTH_LENGTH)
  {
    if (this->logging > 1)
    cout << "Error: wrong number of bytes for fixed length:" << ret << endl;
    return NULL;
  }
  rbuf[ret] = 0;
  nbytes = ret;
  //
  lmsg = atol(rbuf);
  if (this->logging >= 2)
    cout << "::recv() 1:" << " fixed length=" << lmsg << endl;
  if (lmsg < 0L)
  {
    cout << "Error: received message length not OK" << endl;
    return NULL;
  }
  //
  // receive the second message part
  mbuf = new char[lmsg + 2];
  if (mbuf == 0)
  {
    cout << "Error allocation receive buffer" << endl;
    return NULL;
  }
  //
  nb = this->basic_receive(mbuf, lmsg);
  if (nb == 0L) 
  {
    cout << 
      "Error: timeout during receiving the second message part" << endl;
    delete mbuf;
    return NULL;
  }
  if (nb != lmsg)
  {
    cout << 
      "Error: wrong number of bytes second message part" << endl;
    delete mbuf;
    return NULL;
  }
  mbuf[lmsg] = 0;
  pmsg = new MSGSOCKET(mbuf[0], &mbuf[1], lmsg - 1);
  delete mbuf;
  //
  return pmsg;
}   // end SRVSOCKET::recv()
 
// send the required number of bytes
// returns: the number of bytes that have been sent
//          < 0 if error
long SRVSOCKET::basic_send(char *buf, long lbuf)
{
long m, n, nb, anz, ret;
int sleep_count;

  if (logging >= 2)
    cout << "basic_send(): buffer length=" << lbuf << endl;
  sleep_count = 0;
  m   = 0;
  nb  = 0;
  anz = lbuf;
  while (nb < lbuf)
  {
    /* problems under Windows...
    ret = this->test_socket();
    if (ret)
    {
      // this is a severe error
      if (this->logging)
        cout << "\tSRVSOCKET::basic_send returns Error -1 for socket " 
             << sockfd << endl;
      return -1;
    }
    */
#ifdef WIN32
    ret = ::send(this->rw_sockfd, (char *)(&buf[nb]), anz, 0);
#else
    ret = ::send(this->rw_sockfd, (char *)(&buf[nb]), anz, 0);
#endif
    if (ret == 0)
    {
      if (this->logging)
        cout << "TIMEOUT in SRVSOCKET::basic_send()" << endl;
      sleep_count += this->sleep_unit;
      if (sleep_count >= this->timeout_ms)
        break;
      this->milli_sleep(this->sleep_unit);
      continue;
    }
    if (ret < 0)
    {
      // this is a severe error
      if (this->logging)
        cout << "\tSRVSOCKET::basic_send returns Error -2 for socket " 
             << this->rw_sockfd << endl;
      return -2;
    }
    //
    // success here
    if (logging >= 2)
      cout << "\tbasic_send(): no bytes sent=" << ret << endl;
    nb  += ret;
    anz -= ret;
  }  // end while ...
  if (nb != lbuf)
  {
    if (this->logging)
      cout << "\tSRVSOCKET::basic_send returns Error -3" << endl;
    return -3;
  }
  return nb;
}   // end SRVSOCKET::basic_send()

// send an event via a socket
// returns: the number of bytes sent
int SRVSOCKET::send(MSGSOCKET *pmsg,
                    int wait_flag)
{
int  ret, nb;
long ll;
char *pbuf;
char xbuf[20];
char length[LENGTH_LENGTH + 1];
  //
  // send the length (LENGTH bytes) first
  ll = pmsg->get_message_length();
  sprintf(xbuf, "%c0%dlu", '%', LENGTH_LENGTH);  // --->"%010lu"
  sprintf(length, xbuf, (long)ll);
  nb = 0;
  ret = this->basic_send(length, LENGTH_LENGTH);
  if (ret <= 0)
  {
    if (this->logging)
      cout << "Error: send over socket:" << ret << endl;
    return -1;
  }
  nb += ret;
  if (! pmsg->is_binary())
  {
    // non binary data: field structured
    pbuf = pmsg->flatten();
    if (pbuf == 0)
    {
      if (this->logging)
        cout << "Error: memory allocation in send()"  << endl;
      return -2;
    }
  }
  else
  {
    //binary data
    pbuf = new char[ll];
    pbuf[0] = pmsg->get_operation();
    memcpy(&pbuf[1], pmsg->get_content(), ll - 1);
  }
  ret = this->basic_send(pbuf, ll);
  if (ret < 0)
  {
    if (this->logging)
      cout << "Error: send over socket:" << ret << endl;
    return -3;
  }
  delete pbuf;
  nb += ret;
  return nb;
}   // end SRVSOCKET::send()

//
// Returns: 0  if ok
//          <0 if error
int SRVSOCKET::open(int port_number)
{
struct sockaddr_in server; 
int                optval;
  if (port_number < 0)
  {
    cerr << "Error: port number < 0" << endl;
    return -1;
  }
  //
  // prepare our main (server) socket connection
  this->listen_sockfd = ::socket(AF_INET, SOCK_STREAM, 0);
  if (this->listen_sockfd == INVALID_SOCKET)
  {
    cerr << "socket call fails" << endl;
    return -3;
  }
  if (this->logging)
    cout << "listen socket ok:" << this->listen_sockfd << endl;
  optval = 1;
  ::setsockopt(this->listen_sockfd, IPPROTO_TCP, TCP_NODELAY, 
               (const char *)&optval, (size_t)(sizeof(optval)));
  optval = 1;
  ::setsockopt(this->listen_sockfd, IPPROTO_TCP, SO_REUSEADDR, 
                (const char *)&optval, (size_t)(sizeof(optval)));
  //
  server.sin_port          = htons(port_number);
  server.sin_family        = AF_INET;
  server.sin_addr.s_addr   = INADDR_ANY;
  if (::bind(this->listen_sockfd, (struct sockaddr *)&server, 
                      sizeof(struct sockaddr_in)) < 0)
  {
    cerr << "bind call fails" << endl;
    this->close();
    return -4;
  }
  if (this->logging)
    cout << "bind ok to port " << port_number << endl;

  if (::listen(this->listen_sockfd, 5) < 0)
  {
#ifdef WIN32
    int ret;
    ret = WSAGetLastError();
    cout << "WSA error:" << ret << endl;
#endif
    cerr << "listen call fails" << endl;
    this->close();
    return -5;
  }
  if (this->logging)
    cout << "listen ok" << endl;
  return 0;
}   // end SRVSOCKET::open()

// close the read-write socket only
int SRVSOCKET::close_rwsocket()
{
  if (this->rw_sockfd != INVALID_SOCKET)
  {
#ifdef WIN32
    ::closesocket(this->rw_sockfd);
#else
    ::close(this->rw_sockfd);
#endif
  }
  this->rw_sockfd = INVALID_SOCKET;
  return 0;
}

// close both types of server sockets
int SRVSOCKET::close()
{
  if (this->listen_sockfd != INVALID_SOCKET)
  {
#ifdef WIN32
    ::closesocket(this->listen_sockfd);
#else
    ::close(this->listen_sockfd);
#endif
  }
  this->listen_sockfd = INVALID_SOCKET;
  this->close_rwsocket();
  return 0;
}

// Returns the status of a socket
// Returns -1 if the connection is closed or not yet open
int SRVSOCKET::status()
{
  if (this->listen_sockfd == INVALID_SOCKET)
    return -1;
  if (this->rw_sockfd == INVALID_SOCKET)
    return -1;
  return 0;
}

void SRVSOCKET::setLogging(int log_flag)
{ 
  this->logging = log_flag; 
}

int SRVSOCKET::getLogging()
{
  return this->logging;
}

// set a new timeout value in milli seconds
void SRVSOCKET::set_timeout(long ms)
{
  this->timeout_ms = ms;
}

